/***************************************************************************
 *  @file concat.c
 *
 *  @date   : 05-09-04
 *  @author : Pedro Ortega C. <peortega@dcc.uchile.cl>
 *  Copyright  2004  Pedro Ortega C.
 ****************************************************************************/
/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/******************/
/* Include files. */
/******************/

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "file.h"
#include "commands.h"



/************************/
/* Static declarations. */
/************************/

/**************************/
/* Static implementation. */
/**************************/

/************************************/
/* Public interface implementation. */
/************************************/

int
concat_help(FILE *fp)
{
    fprintf(fp, "\n");
    fprintf(fp, "Concatenates the given sources.\n");
    fprintf(fp, "\n");
    fprintf(fp, "Use: bwtool concat [src1 [src2 .. [srcN]]] <dest>\n");
    fprintf(fp, "\n");
    fprintf(fp, "Note: bwtool tries to maximize the amount of channels to "
                "be copied to <dest>.\n");
    fprintf(fp, "\n");
}


int
concat(int argc, char **argv)
{
    int i;
    int ret;
    int max;

    bwfile *src;
    bwfile *dest;
    bwsample *in;
    bwsample *out;

    assert(argc >= 2);
    assert(argv != NULL);

    // Check if only 3 args were given
    if (argc == 3)
    {
        concat_help(stderr);
        return 0;
    }

    // Check the sources' numbers of channels and get the maximum
    max = 0;
    for (i=2; i<argc-1; ++i)
    {
        // Open the brainwave stream
        src = bwopen(argv[2]);
        if (src == NULL)
        {
            fprintf(stderr, "Couldn't open \"%s\".\n", argv[2]);
            return 1;
        }

        // check its channels
        if (max < src->nchan)
            max = src->nchan;

        // close the file
        bwclose(src);
    }

    // Open the output brainwave stream
    dest = bwcreate(argv[argc-1], max);
    if (dest == NULL)
    {
        fprintf(stderr, "Couldn't open \"%s\".\n", argv[3]);
        bwclose(src);
        return 1;
    }
    
    // Warn if destiny file can't take that much channels
    if (dest->nchan < max)
    {
        fprintf(stderr, "Warning: Output file \"%s\" will have %d channels.\n",
                dest->name, dest->nchan);
        max = dest->nchan;
    }

    // create output sample structure
    out = bwsample_new(dest->nchan);

    // cycle through all input files
    for (i=2; i<argc-1; ++i)
    {
        int j;

        // open brainwave file
        src = bwopen(argv[i]);
        if (src == NULL)
        {
            fprintf(stderr, "Couldn't open \"%s\".\n", argv[2]);
            return 1;
        }
        
        // create input sample structure
        in  = bwsample_new(src->nchan);

        // check its channels
        if (src->nchan < max)
        {
            fprintf(stderr, "Warning: File \"%s\" has less channels than "
                    "required. The lacking channels will be set at 512.\n",
                    src->name);
            for (j=0; j<dest->nchan; ++j)
                out->chan[j] = 512;
        }

        // copy input to output
        while (bwread(src, in) != EOF)
        {
            for (j=0; j<max; ++j)
                out->chan[j] = in->chan[j];
            bwwrite(dest, out);
        }

        // close the file and destroy input sample structure
        bwclose(src);
        bwsample_destroy(in);
    }

    // close
    bwclose(dest);
    bwsample_destroy(out);

    return 0;
}



