/***************************************************************************
 *  @file convert.c
 *
 *  @date   : 05-09-04
 *  @author : Pedro Ortega C. <peortega@dcc.uchile.cl>
 *  Copyright  2004  Pedro Ortega C.
 ****************************************************************************/
/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/******************/
/* Include files. */
/******************/

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "file.h"
#include "commands.h"



/************************/
/* Static declarations. */
/************************/

/**************************/
/* Static implementation. */
/**************************/

/************************************/
/* Public interface implementation. */
/************************************/

int
convert_help(FILE *fp)
{
    fprintf(fp, "\n");
    fprintf(fp, "Convert the given source file into another file.\n");
    fprintf(fp, "\n");
    fprintf(fp, "Use: bwtool convert <src> <dest>\n");
    fprintf(fp, "\n");
    fprintf(fp, "Note: Not all formats support a variable number of channels.\n"
                "If the format of the destination file requires more channels "
                "than those that\n"
                "are available, then bwtool will create them. If the target "
                "format doesn't allow\n"
                "to contain all source channels, then only the first ones "
                "will be converted\n"
                "to <dest>.\n");
    fprintf(fp, "\n");
    
    return 0;
}

int
convert(int argc, char **argv)
{
    int max;
    bwfile *src;
    bwfile *dest;
    bwsample *in;
    bwsample *out;

    // check arguments
    if (argc != 4)
    {
        convert_help(stderr);
        return 1;
    }

    // open brainwave streams
    src  = bwopen(argv[2]);
    if (src == NULL)
    {
        fprintf(stderr, "Couldn't open \"%s\".\n", argv[2]);
        return 1;
    }

    dest = bwcreate(argv[3], src->nchan);
    if (dest == NULL)
    {
        fprintf(stderr, "Couldn't open \"%s\".\n", argv[3]);
        bwclose(src);
        return 1;
    }

    // build sample structures
    in  = bwsample_new(src->nchan);
    out = bwsample_new(dest->nchan);

    // check channel number
    max = src->nchan;
    if (src->nchan < dest->nchan)
    {
        int i;
        fprintf(stderr, "Warning: Lacking channels will be filled with 512.\n");

        for (i=0; i<dest->nchan; ++i)
            out->chan[i] = 512;

        max = src->nchan;
    }
    if (src->nchan > dest->nchan)
    {
        fprintf(stderr, "Warning: Only the first %d "
                "channels will be converted.\n", dest->nchan);

        max = dest->nchan;
    }

    // convert
    while (bwread(src, in) != EOF)
    {
        int i;
        for (i=0; i<max; ++i)
            out->chan[i] = in->chan[i];
        bwwrite(dest, out);
    }

    // close
    bwclose(src);
    bwclose(dest);
    bwsample_destroy(in);
    bwsample_destroy(out);

    return 0;
}


