/***************************************************************************
 *  @file file.c
 *
 *  @date   : 03-09-04
 *  @author : Pedro Ortega C. <peortega@dcc.uchile.cl>,
 *            Some code snippets from Jim Peters <http://uazu.net>
 *  Copyright  2004  Pedro Ortega C.
 ****************************************************************************/
/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/******************/
/* Include files. */
/******************/

#include <assert.h>
#include <string.h>
#include "file.h"



/***********************/
/* Static definitions. */
/***********************/

static char *
bwfile_get_type(const char *name);

static int
bwfile_setup_file_type(bwfile *f, const char *name);

static bwfile*
bwalloc(const char *name, int mode);



static int bwfile_read_p2(bwfile *f, bwsample *s);
static int bwfile_read_p3(bwfile *f, bwsample *s);
static int bwfile_read_bbay(bwfile *f, bwsample *s);
static int bwfile_read_txt(bwfile *f, bwsample *s);
static int bwfile_read_bin(bwfile *f, bwsample *s);
static int bwfile_read_edf(bwfile *f, bwsample *s);

static int bwfile_write_p2(bwfile *f, bwsample *s);
static int bwfile_write_p3(bwfile *f, bwsample *s);
static int bwfile_write_bbay(bwfile *f, bwsample *s);
static int bwfile_write_txt(bwfile *f, bwsample *s);
static int bwfile_write_bin(bwfile *f, bwsample *s);
static int bwfile_write_edf(bwfile *f, bwsample *s);

static int bwfile_setup_p2(bwfile *f);
static int bwfile_setup_p3(bwfile *f);
static int bwfile_setup_bbay(bwfile *f);
static int bwfile_setup_txt(bwfile *f);
static int bwfile_setup_bin(bwfile *f);
static int bwfile_setup_edf(bwfile *f);

static int bwfile_create_p2(bwfile *f, int nchan);
static int bwfile_create_p3(bwfile *f, int nchan);
static int bwfile_create_bbay(bwfile *f, int nchan);
static int bwfile_create_txt(bwfile *f, int nchan);
static int bwfile_create_bin(bwfile *f, int nchan);
static int bwfile_create_edf(bwfile *f, int nchan);



/**************************/
/* Static implementation. */
/**************************/

/**
 * @brief Look at the filename extension and return a string with the format.
 *
 * This functions takes a filename \arg name, looks at its extension,
 * and decides its format. Then, it returns a string containing the
 * format name.
 *
 * @param name The filename to be inspected.
 * @return Returns a new string or NULL if failed.
 */
static char *
bwfile_get_type(const char *name)
{
    int len, i, j;
    char *ext;
    char *type;
    
    assert(name != NULL);
    
    // get length of file name
    len = strlen(name);
    
    // search the file name extension
    ext = (char *) &(name[len - 1]);
    while((name < ext) && *ext != '.')
        ext--;

    // check if extension was found
    if (ext == name)
        return NULL;

    // Extension found, check type.
    if (!strcasecmp(ext, ".p2"))
    {
        type = strdup("p2");
    }
    else if (!strcasecmp(ext, ".p3"))
    {
        type = strdup("p3");
    }
    else if (!strcasecmp(ext, ".bbay") || !strcasecmp(ext, ".arc"))
    {
        type = strdup("bbay");
    }
    else if (!strcasecmp(ext, ".text") || !strcasecmp(ext, ".txt"))
    {
        type = strdup("text");
    }
    else if (!strcasecmp(ext, ".binary") || !strcasecmp(ext, ".bin"))
    {
        type = strdup("bin");
    }
    else if (!strcasecmp(ext, ".edf"))
    {
        type = strdup("edf");
    }
    else
        return NULL;

    return type;
}

/**
 * @brief Sets the file format specific callbacks.
 *
 * This functions takes a blank \ref bwfile \arg f and a format type
 * string \arg type, and sets the format specific fields of \arg f.
 *
 * @param name The filename to be inspected.
 * @return Returns a string or NULL if failed.
 */
static int
bwfile_setup_file_type(bwfile *f, const char *name)
{
    assert(f != NULL);
    assert(name != NULL);

    // try to deduce file type from file extension.
    f->type = bwfile_get_type(name);

    if (!strcmp(f->type, "p2"))
    {
        f->is_text = 0;
        f->read   = bwfile_read_p2;
        f->write  = bwfile_write_p2;
        f->setup  = bwfile_setup_p2;
        f->create = bwfile_create_p2;
    }
    else if (!strcmp(f->type, "p3"))
    {
        f->is_text = 0;
        f->read   = bwfile_read_p3;
        f->write  = bwfile_write_p3;
        f->setup  = bwfile_setup_p3;
        f->create = bwfile_create_p3;
    }
    else if (!strcmp(f->type, "bbay"))
    {
        f->is_text = 0;
        f->read   = bwfile_read_bbay;
        f->write  = bwfile_write_bbay;
        f->setup  = bwfile_setup_bbay;
        f->create = bwfile_create_bbay;
    }
    else if (!strcmp(f->type, "text"))
    {
        f->is_text = 1;
        f->read   = bwfile_read_txt;
        f->write  = bwfile_write_txt;
        f->setup  = bwfile_setup_txt;
        f->create = bwfile_create_txt;
    }
    else if (!strcmp(f->type, "bin"))
    {
        f->is_text = 0;
        f->read   = bwfile_read_bin;
        f->write  = bwfile_write_bin;
        f->setup  = bwfile_setup_bin;
        f->create = bwfile_create_bin;
    }
    else if (!strcmp(f->type, "edf"))
    {
        f->is_text = 1;
        f->read   = bwfile_read_edf;
        f->write  = bwfile_write_edf;
        f->setup  = bwfile_setup_edf;
        f->create = bwfile_create_edf;
    }
    else
        return 1;

    return 0;
}

/**
 * @brief Alloc a bwfile and open it.
 *
 * This functions opens the file \arg name. The function deduces the
 * format type using the filename's extension. The \arg mode should
 * be \code BW_MODE_READ or \code BW_MODE_WRITE.
 *
 * @param name The filename to be opened.
 * @return Returns a bwfile pointer or NULL if failed.
 */
static bwfile*
bwalloc(const char *name, int mode)
{
    int ret;
    bwfile *f;

    assert(name != NULL);
    assert(mode == BW_MODE_READ || mode == BW_MODE_WRITE);

    // create blank bwfile structure
    f = (bwfile *) malloc(sizeof(bwfile));
    if (f == NULL)
    {
        fprintf(stderr, "Couldn't alloc bwfile!\n");
        return NULL;
    }
    f->fp      = NULL;
    f->name    = strdup(name);
    f->type    = NULL;
    f->is_text = 0;
    f->pos     = 0;
    f->counter = -1;
    f->nchan   = 0;
    f->read    = NULL;
    f->write   = NULL;
    f->setup   = NULL;
    f->create  = NULL;

    
    // setup brainwave file type
    ret = bwfile_setup_file_type(f, name);
    if (ret != 0)
    {
        fprintf(stderr, "Error opening \"%s\": Unknown file format!\n", name);
        bwclose(f);
        return NULL;
    }
    
    // open file
    if (mode == BW_MODE_READ && f->is_text)
        f->fp = fopen(name, "r");
    else if (mode == BW_MODE_READ && !f->is_text)
        f->fp = fopen(name, "rb");
    else if (mode == BW_MODE_WRITE && f->is_text)
        f->fp = fopen(name, "w");
    else if (mode == BW_MODE_WRITE && !f->is_text)
        f->fp = fopen(name, "wb");

    if (f->fp == NULL)
    {
        fprintf(stderr, "Error opening \"%s\"!\n", name);
        bwclose(f);
        return NULL;
    }

    return f;
}


/********************************************/
/* Static implementation: reading routines. */
/********************************************/

static int
bwfile_read_p2(bwfile *f, bwsample *s)
{
    unsigned char buf[15];
    int ch1, ch2;
    int cnt, a;

    assert(f != NULL);
    assert(s != NULL);
    assert(f->nchan == 6);
    assert(s->nchan == 6);

    // First scan forwards to the start-mark
    ch1 = fgetc(f->fp);
    cnt = 0;
    if (ch1 == EOF) return EOF;
    while (1)
    {
        ch2 = fgetc(f->fp);
        if (ch2 == EOF) return EOF;
        if (ch1 == 0xa5 && ch2 == 0x5a) break;
        s->err = 1;  // Mark sync error
        ch1 = ch2;
        if (++cnt > 100)
            fprintf(stderr, "No start mark found in 100 bytes of data stream");
    }

    // Read the rest of the packet
    if (1 != fread(buf, 15, 1, f->fp)) return 1;

    // Check the version number
    if (2 != buf[0])
        printf("WARNING: version %d not supported\n", buf[0]);

    // Check the counter.  Note this only checks within the current
    // run of data -- it doesn't check continuity from the last
    // block read.  Also, we could perhaps insert big runs of zeros
    // for gaps in the count, but this does not seem worth it.
    if (f->counter >= 0 && f->counter != buf[1])
        s->err = 1;
    f->counter = 255 & (buf[1] + 1);

    // Convert the data.  Bad data gives error marks
    for (a = 0; a<6; a++)
    {
        int val = (buf[2*a+2]<<8) + buf[2*a+3];
        if (val < 0 || val >= 1024)
        {
            s->err = 1;
            val = 512;
        }
        s->chan[a] = val;
    }

    return 0;
}

static int
bwfile_read_p3(bwfile *f, bwsample *s)
{
    unsigned char buf[64];
    int plen;
    int p_cnt;
    int p_aux;
  
    int ch;
    int cnt, a;

    assert(f != NULL);
    assert(s != NULL);
    assert(f->nchan == 6);
    assert(s->nchan == 6);

    // Read a whole packet in
    plen = 0;
    while (1)
    {
        ch = fgetc(f->fp);
        if (ch == EOF) return EOF;
        buf[plen++]= ch;
        if (ch & 128) break;
        if (plen == 64)
        {  // Extra-long packet (bad data)
            memmove(buf, buf+32, 32);
            plen = 32;
        }
    }

    // Bad packet
    if (plen != 5 && plen != 8 && plen != 11)
    {
        s->err = 1;
        for (a = 0; a<10; a++)
            s->chan[a] = 0;
        return 1;
    }

    // Decode packet
    p_cnt = buf[0] >> 1;

    // Check the counter.  Note this only checks within the current
    // run of data -- it doesn't check continuity from the last
    // block read.
    if (f->counter >= 0 && f->counter != p_cnt)
        s->err = 1;
    f->counter = 63 & (p_cnt + 1);
      
    // Decode channel data
    for (a = 0; a<6; a++)
    {
        int off = 2 + (a>>1) * 3;
        int val = ((off >= plen) ? 512 :
            !(a&1) ? buf[off] + ((buf[off+2] & 0x70) << 3) :
            buf[off+1] + ((buf[off+2] & 0x7) << 7));
        s->chan[a] = val;
    }

    return 0;
}

static int
bwfile_read_bbay(bwfile *f, bwsample *s)
{
    // Empty for now.
    return 0;
}

static int
bwfile_read_txt(bwfile *f, bwsample *s)
{
    int i;
    int c, pos;
    char word[64];
    
    assert(f != NULL);
    assert(s != NULL);
    assert(f->nchan == s->nchan);
    assert(f->nchan < BW_MAX_CHAN);

    for (i=0; i<s->nchan; ++i)
    {
        // skip blank space
        c  = fgetc(f->fp);
        while (c == ' ' || c == '\t')
            c = fgetc(f->fp);
        
        // get chars to build number
        pos = 0;
        while (isdigit(c))
        {
            word[pos++] = c;
            c = fgetc(f->fp);
        }
    
        // build number
        if (pos == 0)
        {
            int j;
            s->err = 1;
            for (j=0; j<s->nchan; ++j)
                s->chan[j] = 512;
            break;
        }
        s->chan[i] = atoi(word);
    }
    
    // get to the next line
    while (c != '\n' && c != EOF)
    {
        c = fgetc(f->fp);
        
        if (c != ' ' || c != '\t')
        {
            int j;
            s->err = 1;
            for (j=0; j<s->nchan; ++j)
                s->chan[j] = 512;
        }
    }

    // if EOF reached, return EOF
    if (c == EOF)
        return EOF;

    return 0;
}

static int
bwfile_read_bin(bwfile *f, bwsample *s)
{
    int i;
    int h, l;

    assert(f != NULL);
    assert(s != NULL);
    assert(f->nchan == s->nchan);
    assert(f->nchan < BW_MAX_CHAN);

    // read the samples
    for (i=0; i<s->nchan; ++i)
    {
        h = fgetc(f->fp);
        if (h == EOF) break;
        l = fgetc(f->fp);
        if (l == EOF) break;
        s->chan[i] = (h << 8) + l;
    }
    
    // mark error if data couldn't be read
    if (i < s->nchan)
    {
        int j;
        s->err = 1;
        for (j=0; j<s->nchan; ++j)
            s->chan[j] = 512;
        if (h == EOF || l == EOF)
            return EOF;
        else
            return 1;
    }

    return 0;
}

static int
bwfile_read_edf(bwfile *f, bwsample *s)
{
    // Empty for now.
    return 0;
}



/********************************************/
/* Static implementation: writing routines. */
/********************************************/

static int
bwfile_write_p2(bwfile *f, bwsample *s)
{
    int i;
    
    assert(f != NULL);
    assert(s != NULL);
    assert(f->nchan == 6);
    assert(s->nchan == 6);
    
    // Start-of-packet
    fputc(0xa5, f->fp);
    fputc(0x5a, f->fp);
    
    // Version number
    fputc(2, f->fp);

    // Packet counter
    f->counter = (f->counter + 1) % 255;
    fputc(f->counter, f->fp);

    // channel data
    for (i=0; i<6; ++i)
    {
        int h, l;
        h = (s->chan[i] & 0xff00) >> 8;
        l = s->chan[i] & 0xff;
        
        fputc(h, f->fp);
        fputc(l, f->fp);
    }
    
    // switch settings
    fputc(0x0f, f->fp);
    
    return 0;
}

static int
bwfile_write_p3(bwfile *f, bwsample *s)
{
    int a;
    int aux, h1, h2;
    int ch1, ch2, ch3;
    
    assert(f != NULL);
    assert(s != NULL);
    assert(f->nchan == 6);
    assert(s->nchan == 6);

    // print header
    f->counter = (f->counter + 1) % 64;
    h1 = (f->counter << 1);
    if ((f->counter & 0x7) == 4)
        aux = 0xf;
    else
        aux = 0;
    h1 = h1 | (aux >> 7);
    h2 = aux & 0x7f;
    fputc(h1, f->fp);
    fputc(h2, f->fp);
    
    // Encode channel data
    for (a = 0; a<3; a++)
    {
        ch1 = s->chan[2*a] & 0x7f;
        ch2 = s->chan[2*a+1] & 0x7f;
        ch3 = ((s->chan[2*a] & 0x380) >> 3) | ((s->chan[2*a+1] & 0x380) >> 7);
        if (a==2) ch3 |= 0x80;
        fputc(ch1, f->fp);
        fputc(ch2, f->fp);
        fputc(ch3, f->fp);
    }

    return 0;
}

static int
bwfile_write_bbay(bwfile *f, bwsample *s)
{
    // nothing for now.

    return 0;
}

static int
bwfile_write_txt(bwfile *f, bwsample *s)
{
    int i;

    assert(f != NULL);
    assert(s != NULL);
    assert(f->nchan == s->nchan);

    // print line
    for (i=0; i<s->nchan-1; ++i)
        fprintf(f->fp, "%d ", s->chan[i]);
    fprintf(f->fp, "%d\n", s->chan[s->nchan-1]);

    return 0;
}

static int
bwfile_write_bin(bwfile *f, bwsample *s)
{
    int i;

    assert(f != NULL);
    assert(s != NULL);
    assert(f->nchan == s->nchan);

    // print channel data
    for (i=0; i<s->nchan; ++i)
    {
        int h, l;
        h = (s->chan[i] & 0xff00) >> 8;
        l = (s->chan[i] & 0xff);
        fputc(h, f->fp);
        fputc(l, f->fp);
    }

    return 0;
}

static int
bwfile_write_edf(bwfile *f, bwsample *s)
{
    // nothing for now.

    return 0;
}



/******************************************/
/* Static implementation: setup routines. */
/******************************************/

static int
bwfile_setup_p2(bwfile *f)
{
    assert(f != NULL);

    f->counter = -1;
    f->nchan = 6;

    return 0;
}

static int
bwfile_setup_p3(bwfile *f)
{
    assert(f != NULL);

    f->counter = -1;
    f->nchan = 6;

    return 0;
}

static int
bwfile_setup_bbay(bwfile *f)
{
    return 0;
}

static int
bwfile_setup_txt(bwfile *f)
{
    int ret;
    int nchan;

    assert(f != NULL);

    f->counter = 0;

    // count amount of channels
    ret = fscanf(f->fp, "channels: %d \n", &nchan);
    if (ret != 1)
    {
        fprintf(stderr,
                "Amount of channels expected in file \"%s\" at line 1!\n",
                f->name);
        return 1;
    }

    f->nchan = nchan;

    return 0;
}

static int
bwfile_setup_bin(bwfile *f)
{
    int h, l, nchan;

    assert(f != NULL);

    f->counter = 0;

    // count amount of channels
    h = fgetc(f->fp);
    if (h == EOF)
    {
        fprintf(stderr, "File \"%s\" is empty!\n", f->name);
        return 1;
    }

    l = fgetc(f->fp);
    if (h == EOF)
    {
        fprintf(stderr, "Error in \"%s\": Channel number expected!\n", f->name);
        return 1;
    }

    f->nchan = (h << 8) + l;

    return 0;
}

static int
bwfile_setup_edf(bwfile *f)
{
    // not implemented yet
    
    return 0;
}



/*******************************************/
/* Static implementation: create routines. */
/*******************************************/

static int
bwfile_create_p2(bwfile *f, int nchan)
{
    assert(f != NULL);

    if (nchan != 6)
        fprintf(stderr, "Warning: P2 type supports only 6 channels.\n");
    
    // set amount fo channels
    f->nchan = 6;
    
    return 0;
}

static int
bwfile_create_p3(bwfile *f, int nchan)
{
    assert(f != NULL);

    if (nchan != 6)
        fprintf(stderr, "Warning: P3 type supports only 6 channels.\n");
    
    // set amount fo channels
    f->nchan = 6;
    
    return 0;
}

static int
bwfile_create_bbay(bwfile *f, int nchan)
{
    return 0;
}

static int
bwfile_create_txt(bwfile *f, int nchan)
{
    assert(f != NULL);

    if (nchan < 0)
    {
        fprintf(stderr, "Channel number should be strictly positive.\n");
        return 1;
    }
    if (nchan > BW_MAX_CHAN)
    {
        fprintf(stderr, "No more than %d channels supported.\n", BW_MAX_CHAN);
        return 1;
    }

    // set amount fo channels
    f->nchan = nchan;

    // print amount of channels
    fprintf(f->fp, "channels: %d \n", nchan);

    return 0;
}

static int
bwfile_create_bin(bwfile *f, int nchan)
{
    assert(f != NULL);

    if (nchan < 0)
    {
        fprintf(stderr, "Channel number should be strictly positive.\n");
        return 1;
    }
    if (nchan > BW_MAX_CHAN)
    {
        fprintf(stderr, "No more than %d channels supported.\n", BW_MAX_CHAN);
        return 1;
    }

    // set amount fo channels
    f->nchan = nchan;
    
    // print amount of channels
    fputc((nchan & 0xff00) >> 8, f->fp);
    fputc(nchan & 0xff, f->fp);

    return 0;
}

static int
bwfile_create_edf(bwfile *f, int nchan)
{
    return 0;
}



/************************************/
/* Public interface implementation. */
/************************************/

/**
 * @brief Opens a file with brainwave data.
 *
 * This function opens a file containing brainwave data. The file format
 * will be deduced from the file extension.
 *
 * @param  name Filename.
 * @return Returns a pointer to a new brainwave file pointer.
 */
bwfile*
bwopen(const char *name)
{
    int ret;
    bwfile *f;
    FILE *fp;
    
    assert(name != NULL);
    
    // open file
    f = bwalloc(name, BW_MODE_READ);
    if (f == NULL)
        return NULL;

    // setup
    ret = f->setup(f);
    if (ret != 0)
    {
        bwclose(f);
        return NULL;
    }
    
    return f;
}

/**
 * @brief Create a new file with brainwave data.
 *
 * This function creates a new file containing brainwave data. The file format
 * will be selected looking at the file extension.
 *
 * @param  name Filename.
 * @return Returns a pointer to a new brainwave file pointer.
 */
bwfile*
bwcreate(const char *name, int nchan)
{
    int ret;
    bwfile *f;

    // open file
    f = bwalloc(name, BW_MODE_WRITE);
    if (f == NULL)
        return NULL;

    // create
    ret = f->create(f, nchan);
    if (ret != 0)
    {
        if (f->name)
            free(f->name);
        free(f);
        return NULL;
    }

    return f;
}

/**
 * @brief Closes a file.
 *
 * This function closes the pointed brainwave file.
 *
 * @param f A pointer to a brainwave file.
 * @return Returns 0 if OK.
 */
int
bwclose(bwfile *f)
{
    assert(f != NULL);
    
    if (f->fp)
        fclose(f->fp);
    if (f->name)
        free(f->name);
    free(f);
    
    return 0;
}

/**
 * @brief Reads the next sample.
 *
 * This function reads the next sample from the brainwave file. The
 * function fills the (previously allocated) struct pointed by \arg s
 * with the data.
 *
 * @param f The pointer to the brainwave file.
 * @param s A pointer to a \ref bwsample struct where this function should
 *          place the sample data.
 * @return Returns 0 if OK.
 */
int
bwread(bwfile *f, bwsample *s)
{
    assert(f != NULL);
    assert(s != NULL);
    
    f->pos++;
    s->err = 0;
    return f->read(f, s);
}

/**
 * @brief Writes a sample to the file.
 *
 * This function writes a sample to the brainwave file pointed by \arg f.
 * The specific format is given by the configuration in the \ref bwfile
 * structure.
 *
 * @param f A pointer to a brainwave file.
 * @param s A pointer to the sample data.
 * @return Returns 0 if OK.
 */
int
bwwrite(bwfile *f, bwsample *s)
{
    assert(f != NULL);
    assert(s != NULL);
    
    f->pos++;

    return f->write(f, s);
}

bwsample *
bwsample_new(int nchan)
{
    bwsample *s;
    
    assert(nchan > 0);
    
    s = (bwsample *) malloc(sizeof(*s));
    if (s == NULL)
    {
        fprintf(stderr, "Couldn't alloc memory for bwsample structure.\n");
        return NULL;
    }

    s->n     = 0;
    s->nchan = nchan;
    s->chan  = (int *) malloc(sizeof(int) * nchan);
    if (s->chan == NULL)
    {
        fprintf(stderr, "Couldn't alloc memory for bwsample structure.\n");
        bwsample_destroy(s);
        return NULL;
    }
    
    s->err   = 0;
    
    return s;
}

int
bwsample_destroy(bwsample *s)
{
    if (s == NULL)
        return 0;
    if (s->chan)
        free(s->chan);
    free(s);
    
    return 0;
}

int
bwsample_resize(bwsample *s, int nchan)
{
    assert(s != NULL);
    assert(s->chan != NULL);
    
    if (nchan <= 0)
    {
        fprintf(stderr,
            "Amount of channels should be greater or equal to 1.\n");
        return 1;
    }
    if (nchan > BW_MAX_CHAN)
    {
        fprintf(stderr,
            "Maximum amount of allowed channels is %d.\n", BW_MAX_CHAN);
        return 1;
    }
    
    // delete old array and alloc a new one
    free(s->chan);
    s->chan = (int *) malloc(sizeof(int) * nchan);
    if (s->chan == NULL)
    {
        fprintf(stderr, "Failed to resize the sample size!\n");
        return 1;
    }
    
    // change settings
    s->nchan = nchan;
    
    return 0;
}




