/***************************************************************************
 *  @file main.c
 *
 *  @date   : 04-09-04
 *  @author : Pedro Ortega C. <peortega@dcc.uchile.cl>
 *  Copyright  2004  Pedro Ortega C.
 ****************************************************************************/
/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>

#include "main.h"
#include "file.h"
#include "commands.h"



/*************************/
/* Command registration. */
/*************************/

bwcommand commands[] =
{
    {
        "convert", "Convert format type.",
        "<src> <dest>",
        convert, convert_help
    },
    {
        "select", "Select intervals.",
        "[int1 [int2 .. [intN]]] <src> <dest>",
        select, select_help
    },
    {
        "select-channel", "Select channels.",
        "[ch1 [ch2 .. [chN]]] <src> <dest>",
        select_channel, select_channel_help
    },
    {
        "concat", "Concatenate files.",
        "[src1 [src2 .. [srcN]]] <dest>",
        concat, concat_help
    },
    {
        "merge-channel", "Merge channels from files.",
        "[src1 [src2 .. [srcN]]] <dest>",
        merge_channel, merge_channel_help
    },
    {
        "display-info", "Display info.",
        "<file>",
        display_info, display_info_help
    },
    {
        NULL, NULL,
        NULL,
        NULL, NULL
    }
};



/*****************/
/* Declarations. */
/*****************/

int help(FILE *fp);
int help_cmd(FILE *fp, const char *cmd);
int main(int argc, char **argv);



/**
 * @brief Display basic help info.
 *
 * This function prints the execution modes.
 *
 * @param fp The stream to print the info.
 * @return Returns 0 if OK.
 */
int
help(FILE *fp)
{
    bwcommand *ptr;
    
    fprintf(fp, "\n");
    fprintf(fp, "%s %s (c) 2004 by Pedro Ortega <peortega@dcc.uchile.cl>\n",
                BW_PROGNAME, BW_VERSION);
    fprintf(fp, "\n");

    // print command lines
    fprintf(fp, "Use:\n");
    ptr = commands;
    while (ptr->name != NULL)
    {
        fprintf(fp, "%s %s %s\n",
                BW_PROGNAME, ptr->name, ptr->short_args_format);
        ptr++;
    }
    fprintf(fp, "\n");

    // print short command descriptions
    fprintf(fp, "Short description:\n");
    ptr = commands;
    while (ptr->name != NULL)
    {
        fprintf(fp, "%s: %s\n", ptr->name, ptr->short_description);
        ptr++;
    }
    fprintf(fp, "\n");
    fprintf(fp, "Use \"%s help <command-name>\" for specific help.\n",
                BW_PROGNAME);
    fprintf(fp, "\n");

    return 0;
}


/**
 * @brief Display basic help info on a specific command.
 *
 * This function prints the command's specific help. If the given
 * command is \code NULL, then print just general help.
 *
 * @param fp The stream to print the info.
 * @return Returns 0 if OK.
 */
int
help_cmd(FILE *fp, const char *cmd)
{
    int ret;
    bwcommand *ptr;

    // Check if cmd is NULL
    if (cmd == NULL)
    {
        help(fp);
        return 0;
    }

    // Check which command is being invoked.
    ptr = commands;
    while (ptr->name != NULL && strcmp(cmd, ptr->name) != 0)
        ptr++;

    if (ptr->name != NULL && strcmp(cmd, ptr->name) == 0)
        ret = ptr->help(fp);
    else
    {
        help(stderr);
        fprintf(stderr, "Unknown command: \"%s\"\n", cmd);
        fprintf(stderr, "\n");
        return 1;
    }

    return ret;
}



int
main(int argc, char **argv)
{
    int ret;
    bwcommand *ptr;

    // check argument number
    if (argc < 2)
    {
        help(stderr);
        exit(0);
    }

    // Check which command is being invoked.
    ptr = commands;
    while (ptr->name != NULL && strcmp(argv[1], ptr->name) != 0)
        ptr++;

    if (ptr->name != NULL && strcmp(argv[1], ptr->name) == 0)
        ret = ptr->exec(argc, argv);
    else if (!strcmp(argv[1], "help"))
        ret = help_cmd(stdout, argv[2]);
    else
    {
        help(stderr);
        fprintf(stderr, "Unknown command: \"%s\"\n", argv[1]);
        fprintf(stderr, "\n");
        exit(1);
    }

    return ret;
}

