/***************************************************************************
 *  @file merge_channel.c
 *
 *  @date   : 05-09-04
 *  @author : Pedro Ortega C. <peortega@dcc.uchile.cl>
 *  Copyright  2004  Pedro Ortega C.
 ****************************************************************************/
/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/******************/
/* Include files. */
/******************/

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "file.h"
#include "commands.h"



/************************/
/* Static declarations. */
/************************/



/**************************/
/* Static implementation. */
/**************************/



/************************************/
/* Public interface implementation. */
/************************************/

int
merge_channel_help(FILE *fp)
{
    fprintf(fp, "\n");
    fprintf(fp, "Packs all the sources' channels into a single file.\n");
    fprintf(fp, "\n");
    fprintf(fp, "Use: bwtool merge-channel [src1 [src2 .. [srcN]]] <dest>\n");
    fprintf(fp, "\n");
    fprintf(fp, "Note: bwtool tries to maximize the amount of channels to "
                "be copied to \n"
                "<dest>. If the sources have more channels than supported "
                "by the target\n"
                "format, then the rest is dropped. If the target requires "
                "more channels \n"
                "than available, then bwtool creates the rest, setting "
                "their values to 512.\n");
    fprintf(fp, "\n");

    return 0;
}

int
merge_channel(int argc, char **argv)
{
    int i;
    int j;
    int opened;
    int ret;
    int max;

    bwfile *src[BW_MAX_FILES];
    bwfile *dest;
    bwsample *in[BW_MAX_FILES];
    bwsample *out;
    int open[BW_MAX_FILES];

    assert(argc >= 2);
    assert(argv != NULL);

    // Check if only 3 args were given
    if (argc <= 3)
    {
        merge_channel_help(stderr);
        return 0;
    }
    
    // Check if maximum amount of input files have been exceeded
    if (argc-3 > BW_MAX_FILES)
    {
        fprintf(stderr, "Too many input files.\n");
        return 1;
    }

    // Check the source's numbers of channels
    max = 0;
    opened = 0;
    for (i=2; i<argc-1; ++i)
    {
        src[i-2] = bwopen(argv[i]);
        if (src == NULL)
        {
            fprintf(stderr, "Couldn't open \"%s\".\n", argv[i]);
            for (j=2; j<i; ++j)
            {
                bwclose(src[j-2]);
                bwsample_destroy(in[j-2]);
            }
            return 1;
        }
        // create input sample structure
        in[i-2] = bwsample_new(src[i-2]->nchan);

        max += src[i-2]->nchan;

        // mark this file as opened
        open[i-2] = 1;
        opened++;
    }

    // Open the output brainwave stream
    dest = bwcreate(argv[argc-1], max);
    if (dest == NULL)
    {
        fprintf(stderr, "Couldn't open \"%s\".\n", argv[argc-1]);
        for (j=2; j<argc-1; ++j)
        {
            bwclose(src[j-2]);
            bwsample_destroy(in[j-2]);
        }
        return 1;
    }

    // create output sample structure
    out = bwsample_new(dest->nchan);

    // Warn if destiny file can't take that much channels
    if (dest->nchan < max)
    {
        fprintf(stderr, "Warning: Output file \"%s\" will have %d channels.\n",
                dest->name, dest->nchan);
        max = dest->nchan;
    }
    else if (max < dest->nchan)
    {
        fprintf(stderr, "Warning: The lacking channels will be set at 512.\n");
        for (j=0; j<dest->nchan; ++j)
            out->chan[j] = 512;
    }

    // cycle through all the files, and copy channels
    while (opened > 0)
    {
        int k;
        k = 0;
        
        for (i=0; i<argc-3; ++i)
        {
            // if stream has reached the end, then
            // transfer empty channels (i.e. constant value 512)
            if (!open[i])
            {
                for (j=0; j<src[i]->nchan && k<max; ++j, ++k)
                    out->chan[k] = 512;
                continue;
            }
            
            // read next sample from stream
            ret = bwread(src[i], in[i]);

            // check if EOF reached
            if (ret == EOF)
            {
                open[i] = 0;
                opened--;

                // copy empty data
                for (j=0; j<src[i]->nchan && k<max; ++j, ++k)
                    out->chan[k] = 512;
                continue;
            }
            
            // copy samples
            for (j=0; j<src[i]->nchan && k<max; ++j, ++k)
                out->chan[k] = in[i]->chan[j];
        }
        
        // write recolected channels to output stream
        bwwrite(dest, out);
    }
    
    // close all input files and destroy their sample structures
    for (i=0; i<argc-3; ++i)
    {
        bwclose(src[i]);
        bwsample_destroy(in[i]);
    }
    bwclose(dest);
    bwsample_destroy(out);

    return 0;
}


