/***************************************************************************
 *  @file select.c
 *
 *  @date   : 05-09-04
 *  @author : Pedro Ortega C. <peortega@dcc.uchile.cl>
 *  Copyright  2004  Pedro Ortega C.
 ****************************************************************************/
/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/******************/
/* Include files. */
/******************/

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "file.h"
#include "commands.h"



/************************/
/* Static declarations. */
/************************/





/**************************/
/* Static implementation. */
/**************************/





/************************************/
/* Public interface implementation. */
/************************************/

int
select_help(FILE *fp)
{
    fprintf(fp, "\n");
    fprintf(fp, "Selects the given time intervals from the source file.\n");
    fprintf(fp, "\n");
    fprintf(fp, "Use: bwtool select [int1 [int2 .. [intN]]] <src> <dest>\n");
    fprintf(fp, "\n");
    fprintf(fp, "Each time interval has the format \"<start>-<end>\", "
                "where <start> and <end>\n");
    fprintf(fp, "are times expressed in seconds.\n");
    fprintf(fp, "\n");
    fprintf(fp, "Examples:\n");
    fprintf(fp, "bwtool select 10-20 30-40 50-60 input.p2 output.p2\n");
    fprintf(fp, "bwtool select 15.5-40.3 input.bin output.bin\n");
    fprintf(fp, "\n");
    
    return 0;
}

int
select(int argc, char **argv)
{
    int i;
    int j;
    int ret;
    int max;

    bwfile *src;
    bwfile *dest;
    bwsample *in;
    bwsample *out;

    assert(argc >= 2);
    assert(argv != NULL);

    // Check if only 3 args were given
    if (argc <= 3)
    {
        select_help(stderr);
        return 0;
    }

    // Check the source's numbers of channels
    max = 0;
    src = bwopen(argv[argc-2]);
    if (src == NULL)
    {
        fprintf(stderr, "Couldn't open \"%s\".\n", argv[argc-2]);
        return 1;
    }
    max = src->nchan;

    // create input sample structure
    in = bwsample_new(src->nchan);

    // close input file
    bwclose(src);



    // Open the output brainwave stream
    dest = bwcreate(argv[argc-1], max);
    if (dest == NULL)
    {
        fprintf(stderr, "Couldn't open \"%s\".\n", argv[argc-1]);
        bwclose(src);
        return 1;
    }

    // create output sample structure
    out = bwsample_new(dest->nchan);

    // Warn if destiny file can't take that much channels
    if (dest->nchan < max)
    {
        fprintf(stderr, "Warning: Output file \"%s\" will have %d channels.\n",
                dest->name, dest->nchan);
        max = dest->nchan;
    }
    else if (max < dest->nchan)
    {
        fprintf(stderr, "Warning: The lacking channels will be set at 512.\n",
                src->name);
        for (j=0; j<dest->nchan; ++j)
            out->chan[j] = 512;
    }

    // cycle through all given intervals
    for (i=2; i<argc-2; ++i)
    {
        float st, et;
        size_t ss, es;

        // open brainwave file
        src = bwopen(argv[argc-2]);
        if (src == NULL)
        {
            fprintf(stderr, "Couldn't open \"%s\".\n", argv[2]);
            return 1;
        }

        // get the start and the end of the interval
        ret = sscanf(argv[i], "%g-%g", &st, &et);
        if (ret != 2)
        {
            fprintf(stderr, "Warning: Bad interval \"%s\", skipping it.\n",
                    argv[i]);
        }
        ss = sec2pos(st) + 1;
        es = sec2pos(et) + 1;
        if (ss >= es)
        {
            fprintf(stderr, "Warning: Start time occurs after stop time in "
                    "\"%s\", skipping it.\n", argv[i]);
            bwclose(src);
            continue;
        }

        // search starting sample
        ret = bwread(src, in);
        while (ret != EOF && src->pos < ss)
            ret = bwread(src, in);

        // copy samples
        while (ret != EOF && src->pos <= es)
        {
            for (j=0; j<max; ++j)
                out->chan[j] = in->chan[j];
            bwwrite(dest, out);
            ret = bwread(src, in);
        }

        // check if all samples could be copied
        if (ret == EOF && src->pos <= es)
        {
            fprintf(stderr, "Warning: Couldn't copy the whole interval \"%s\": "
                    "End of file reached.\n", argv[i]);
        }

        // close the file and destroy input sample structure
        bwclose(src);
    }

    // close
    bwclose(dest);
    bwsample_destroy(in);
    bwsample_destroy(out);

    return 0;
}

