/***************************************************************************
 *  @file select_channel.c
 *
 *  @date   : 05-09-04
 *  @author : Pedro Ortega C. <peortega@dcc.uchile.cl>
 *  Copyright  2004  Pedro Ortega C.
 ****************************************************************************/
/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/******************/
/* Include files. */
/******************/

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "file.h"
#include "commands.h"



/************************/
/* Static declarations. */
/************************/

/**************************/
/* Static implementation. */
/**************************/

/************************************/
/* Public interface implementation. */
/************************************/

int
select_channel_help(FILE *fp)
{
    fprintf(fp, "\n");
    fprintf(fp, "Selects the given channels from the source file.\n");
    fprintf(fp, "\n");
    fprintf(fp, "Use: bwtool select [ch1 [ch2 .. [chN]]] <src> <dest>\n");
    fprintf(fp, "\n");
    fprintf(fp, "Channels can be specified one by one, or using ranges.\n");
    fprintf(fp, "The channel index starts at 1.\n");
    fprintf(fp, "\n");
    fprintf(fp, "Examples:\n");
    fprintf(fp, "bwtool select_channel 1 3-6 input.p2 output.txt\n");
    fprintf(fp, "bwtool select_channel 10-15 1-5 input.txt output.bin\n");
    fprintf(fp, "\n");

    return 0;
}

int
select_channel(int argc, char **argv)
{
    int i;
    int j;
    int ret;
    int max;

    bwfile *src;
    bwfile *dest;
    bwsample *in;
    bwsample *out;

    int ch[2*BW_MAX_CHAN];
    int chsize;

    assert(argc >= 2);
    assert(argv != NULL);

    // Check if only 4 or less args were given
    if (argc <= 4)
    {
        select_channel_help(stderr);
        return 0;
    }

    // Check the source's numbers of channels
    src = bwopen(argv[argc-2]);
    if (src == NULL)
    {
        fprintf(stderr, "Couldn't open \"%s\".\n", argv[argc-2]);
        return 1;
    }

    // parse input channels
    max = 0;
    chsize  = 0;
    for (i=2; i<argc-2; ++i)
    {
        int sch, ech;
        
        // Get start and end channel
        ret = sscanf(argv[i], "%d-%d", &sch, &ech);
        if (ret != 2)
        {
            ret = sscanf(argv[i], "%d", &sch);
            if (ret != 1)
            {
                fprintf(stderr, "Wrong channel selection \"%s\".\n", argv[i]);
                bwclose(src);
                return 1;
            }
            else
                ech = sch;
        }
        
        // Validate channel interval
        if (sch < 1)
        {
            fprintf(stderr, "Channel index starts at 1.\n");
            bwclose(src);
            return 1;
        }
        if (sch > ech)
        {
            fprintf(stderr, "Start channel index is greater than "
                            "end index in \"%s\".\n", argv[i]);
            bwclose(src);
            return 1;
        }
        if (ech > src->nchan)
        {
            fprintf(stderr, "Channel range exceeds the number of channels "
                            "available in \"%s\".\n", argv[i]);
            bwclose(src);
            return 1;
        }

        // Register interval
        ch[2*(i-2)]     = sch-1;
        ch[2*(i-2) + 1] = ech-1;
        max += (ech - sch + 1);
        chsize++;
    }

    // create input sample structure
    in = bwsample_new(src->nchan);

    // Open the output brainwave stream
    dest = bwcreate(argv[argc-1], max);
    if (dest == NULL)
    {
        fprintf(stderr, "Couldn't open \"%s\".\n", argv[argc-1]);
        bwclose(src);
        return 1;
    }

    // create output sample structure
    out = bwsample_new(dest->nchan);

    // Warn if destiny file can't take that much channels
    if (dest->nchan < max)
    {
        fprintf(stderr, "Warning: Output file \"%s\" will have %d channels.\n",
                dest->name, dest->nchan);
        max = dest->nchan;
    }
    else if (max < dest->nchan)
    {
        fprintf(stderr, "Warning: The lacking channels will be set at 512.\n",
                src->name);
        for (j=0; j<dest->nchan; ++j)
            out->chan[j] = 512;
    }

    // copy all samples
    while ((ret = bwread(src, in)) != EOF)
    {
        int k;
        k = 0;

        for (i=0; i<chsize; ++i)
        {
            int s, e;
            s = ch[2*i];
            e = ch[2*i+1];
            for (j=s; j<=e && k<max; ++j, ++k)
                out->chan[k] = in->chan[j];
        }
        bwwrite(dest, out);
    }

    // close
    bwclose(src);
    bwclose(dest);
    bwsample_destroy(in);
    bwsample_destroy(out);

    return 0;
}




